"""Add build duration [0392807467dc].

Revision ID: 0392807467dc
Revises: 0.75.0
Create Date: 2025-02-28 15:50:06.960529

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0392807467dc"
down_revision = "0.75.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_build", schema=None) as batch_op:
        batch_op.add_column(sa.Column("duration", sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_build", schema=None) as batch_op:
        batch_op.drop_column("duration")

    # ### end Alembic commands ###
