import{o,F as r,p as i}from"./index-D2iSHVZq.js";import{b as u}from"./@tanstack-CSxjHCME.js";async function c({runId:e}){const a=o(i.runs.detail(e)),t=await fetch(a,{method:"DELETE",credentials:"include",headers:{"Content-Type":"application/json"}});if(!t.ok){const s=await t.json().then(n=>n.detail).catch(()=>`Failed to delete run ${e}`);throw new r({status:t.status,statusText:t.statusText,message:s})}return t.json()}function p(e){return u({mutationFn:c,...e})}export{p as u};
