"""Release [0.53.1].

Revision ID: 0.53.1
Revises: a39c4184c8ce
Create Date: 2023-12-21 15:13:14.476779

"""

# revision identifiers, used by Alembic.
revision = "0.53.1"
down_revision = "a39c4184c8ce"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
