"""Release [0.51.0].

Revision ID: 0.51.0
Revises: 8ed03137cacc
Create Date: 2023-12-08 10:55:40.330540

"""

# revision identifiers, used by Alembic.
revision = "0.51.0"
down_revision = "8ed03137cacc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
