import{j as e}from"./@radix-C7hRs6Kx.js";import{B as r,V as n}from"./index-D2iSHVZq.js";import{P as o}from"./ProBadge-Cp4hb1YT.js";import{P as i,a as l,b as c,c as m,d,e as x}from"./ProCta-EYoV9CvK.js";import{C as f}from"./CodeSnippet-CvI6D0wx.js";import{d as p}from"./@react-router-CNP6g_RL.js";import"./@tanstack-CSxjHCME.js";import"./@reactflow-CQi1Z1Wq.js";import"./check-DK77doTf.js";const u="/assets/acp-DOsXjFc7.webp";function h(){const[a]=p(),t=a.get("artifact");function s(){return t?`zenml artifact list --name='contains:${t}'`:"zenml artifact list"}return e.jsxs(r,{className:"flex flex-wrap justify-between p-5",children:[e.jsxs("div",{children:[e.jsx("p",{className:"text-text-xl font-semibold",children:"Staying Open Source? "}),e.jsx("p",{className:"text-theme-text-secondary",children:"No problem! Use this CLI command to list your artifacts"})]}),e.jsx(f,{code:s()})]})}function A(){return e.jsxs("div",{children:[e.jsxs(n,{className:"flex items-center gap-1",children:[e.jsx("h1",{className:"text-display-xs font-semibold",children:"Artifacts"}),e.jsx(o,{})]}),e.jsxs("div",{className:"layout-container space-y-5 py-5",children:[e.jsxs(i,{className:"relative overflow-y-hidden",children:[e.jsxs("div",{className:"w-full max-w-none space-y-5 lg:max-w-[900px]",children:[e.jsx(l,{children:"Advanced Artifact Management Features with ZenML Pro"}),e.jsx(c,{}),e.jsx(m,{features:[{title:"Artifact Control Plane Dashboard",subtitle:"Artifact management and monitoring"},{title:"Enterprise Security",subtitle:"Social SSO, RBAC, and User Management"},{title:"Managed ZenML Server",subtitle:"On your VPC or hosted on our infrastructure"},{title:"Advanced MLOps",subtitle:"CI/CD/CT, Model Control Plane and more"}]}),e.jsx(d,{})]}),e.jsx(x,{className:"absolute translate-x-[30%] translate-y-[15%] scale-[40%]",src:u,alt:"Screenshot of the ZenML Pro Artifact Control Plane"})]}),e.jsx(h,{})]})]})}export{A as default};
