import{j as o}from"./@radix-C7hRs6Kx.js";import{S as r}from"./Tabs-DxQ8PDOD.js";import{R as m}from"./RunsBody-W4WHf-sq.js";import{R as i}from"./RunSelector-BLuBYNJt.js";import{S as p}from"./StackList-CgmN5H-i.js";import{a as s}from"./@react-router-CNP6g_RL.js";import"./index-D2iSHVZq.js";import"./@tanstack-CSxjHCME.js";import"./@reactflow-CQi1Z1Wq.js";import"./chevron-down-A-rmltmI.js";import"./ComponentIcon-ils7uNAk.js";import"./layout-h3cbx8WZ.js";import"./rocket-Cf-B-XOR.js";import"./DisplayDate-CYVBBSgr.js";import"./InlineAvatar-Cfz4WSLK.js";import"./NestedCollapsible-Dor-bi98.js";import"./type-guards-CaeD8wHO.js";import"./CodeSnippet-CvI6D0wx.js";import"./CollapsibleCard-D20FtrzC.js";import"./chevron-right-double-uNWbJT-C.js";import"./check-circle-mvyzYvIW.js";import"./refresh-CM5T3QeU.js";import"./SearchField-D-h6jXyg.js";import"./dots-horizontal-BGRJCPCs.js";import"./index-DR30v9MZ.js";import"./all-pipeline-runs-query-BGASHYtF.js";import"./ExecutionStatus-XrvT2r65.js";import"./trash-BWSZ7NRK.js";import"./DeleteAlertDialog-DVvXt-S6.js";import"./zod-C0xYeTvL.js";import"./index.esm-D7jFlf5N.js";import"./AlertDialogDropdownItem-BG7-Ki1L.js";import"./delete-run-DlSLEl5T.js";import"./plus-COjQg3AG.js";import"./index-CrhdX_qG.js";import"./stack-detail-query-BAcZJrN3.js";import"./components-Br2ezRib.js";import"./NumberBox-D2A7ENHb.js";import"./ComponentBadge-DKw7Gndh.js";import"./Infobox-D9k5TFH4.js";import"./DialogItem-BHWf3sIB.js";function V(){const{componentId:t}=s();return o.jsx("div",{className:"@container",children:o.jsx(r,{isPanel:!1,stacksTabContent:o.jsx(p,{fixedQueryParams:{component_id:t}}),runsTabContent:o.jsx(i,{children:o.jsx(m,{fixedQueryParams:{stack_component:t}})}),componentId:t})})}export{V as default};
