"""Release [0.38.0].

Revision ID: 0.38.0
Revises: 43a86093b60e
Create Date: 2023-04-12 16:20:27.266829

"""

# revision identifiers, used by Alembic.
revision = "0.38.0"
down_revision = "43a86093b60e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
