"""Release [0.33.0].

Revision ID: 0.33.0
Revises: ec0d785ca296
Create Date: 2023-02-03 14:52:48.563695

"""

# revision identifiers, used by Alembic.
revision = "0.33.0"
down_revision = "ec0d785ca296"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
