"""Release [0.46.0].

Revision ID: 0.46.0
Revises: 0.45.6
Create Date: 2023-11-06 10:59:18.298042

"""

# revision identifiers, used by Alembic.
revision = "0.46.0"
down_revision = "0.45.6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
