"""Release [0.55.3].

Revision ID: 0.55.3
Revises: 0.55.2
Create Date: 2024-02-19 16:36:39.784623

"""

# revision identifiers, used by Alembic.
revision = "0.55.3"
down_revision = "0.55.2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
