import{j as e,r}from"./@radix-C7hRs6Kx.js";import{B as o,aP as n,V as l}from"./index-D2iSHVZq.js";import{P as i}from"./ProBadge-Cp4hb1YT.js";import{P as m,a as c,b as d,c as x,d as u,e as p}from"./ProCta-EYoV9CvK.js";import{C as f}from"./CodeSnippet-CvI6D0wx.js";import{d as h}from"./@react-router-CNP6g_RL.js";import"./@tanstack-CSxjHCME.js";import"./@reactflow-CQi1Z1Wq.js";import"./check-DK77doTf.js";const j="/assets/mcp-Cb1aMeoq.webp";function P(){const[a]=h(),t=a.get("model");function s(){return t?`zenml model list --name='contains:${t}'`:"zenml model list"}return e.jsxs(o,{className:"flex flex-wrap justify-between p-5",children:[e.jsxs("div",{children:[e.jsx("p",{className:"text-text-xl font-semibold",children:"Staying Open Source? "}),e.jsx("p",{className:"text-theme-text-secondary",children:"No problem! Use this CLI command to list your models"})]}),e.jsx(f,{code:s()})]})}function A(){const{setTourState:a,tourState:{tourActive:t}}=n();return r.useEffect(()=>{t&&a(s=>({...s,run:!0,stepIndex:s.stepIndex}))},[t]),e.jsxs("div",{children:[e.jsxs(l,{className:"flex items-center gap-1",children:[e.jsx("h1",{className:"text-display-xs font-semibold",children:"Models"}),e.jsx(i,{})]}),e.jsxs("div",{className:"layout-container space-y-5 py-5",children:[e.jsxs(m,{className:"relative overflow-y-hidden",children:[e.jsxs("div",{className:"w-full max-w-none space-y-5 lg:max-w-[900px]",children:[e.jsx(c,{children:"Access Advanced Model Management Features with ZenML Pro"}),e.jsx(d,{}),e.jsx(x,{features:[{title:"Model Control Plane Dashboard",subtitle:"Centralized model management and monitoring"},{title:"Enterprise Security",subtitle:"Social SSO, RBAC, and User Management"},{title:"Managed ZenML Server",subtitle:"On your VPC or hosted on our infrastructure"},{title:"Advanced MLOps",subtitle:"CI/CD/CT, Artifact Control Plane and more"}]}),e.jsx(u,{})]}),e.jsx(p,{className:"absolute translate-x-[30%] translate-y-[15%] scale-[40%]",src:j,alt:"Screenshot of the ZenML Pro Artifact Control Plane"})]}),e.jsx(P,{})]})]})}export{A as default};
