import{j as e}from"./@radix-C7hRs6Kx.js";import{R as o}from"./index-D2iSHVZq.js";function a(){return e.jsx(o,{color:"yellow",rounded:!1,className:"rounded-md",children:e.jsx("span",{className:"font-semibold uppercase",children:"Pro"})})}export{a as P};
