import{j as e}from"./@radix-C7hRs6Kx.js";import{m as a,R as l,h as i,r as d}from"./index-D2iSHVZq.js";import{S as c}from"./check-DK77doTf.js";import{L as m}from"./@react-router-CNP6g_RL.js";function g({className:s,...r}){return e.jsx("div",{className:a("flex items-center overflow-hidden rounded-md border border-warning-100 bg-theme-surface-primary px-7 py-5",s),...r})}function j({className:s,...r}){return e.jsx("h2",{className:a("text-display-xs font-semibold",s),...r})}function w(){return e.jsxs("div",{className:"flex w-fit gap-2 rounded-md bg-[#FFF6EA] px-3 py-1",children:[e.jsx("p",{children:"Instantly access Pro features while keeping your existing setup"}),e.jsx(l,{color:"yellow",rounded:!0,size:"sm",className:"font-semibold uppercase",children:"New"})]})}function b({className:s,features:r,...t}){return e.jsx("ul",{...t,className:a("grid grid-cols-1 gap-x-7 gap-y-5 md:grid-cols-2"),children:r.map((n,o)=>e.jsxs("li",{className:"flex items-start gap-2",children:[e.jsx(u,{}),e.jsxs("div",{children:[e.jsx("p",{className:"font-semibold",children:n.title}),e.jsx("p",{className:"text-text-sm text-theme-text-secondary",children:n.subtitle})]})]},o))})}function N(){return e.jsxs("div",{className:"flex w-fit flex-wrap gap-2",children:[e.jsx(i,{asChild:!0,className:"truncate bg-warning-600 hover:bg-warning-400 active:bg-warning-500 active:ring-warning-50",size:"md",children:e.jsx(m,{to:d.upgrade,children:"Try Pro Features"})}),e.jsx(i,{asChild:!0,intent:"secondary",className:"w-fit truncate",emphasis:"minimal",size:"md",children:e.jsx("a",{href:"https://www.zenml.io/pro",target:"_blank",rel:"noopener noreferrer",children:"Learn More"})})]})}function y({className:s,...r}){return e.jsx("img",{className:a("hidden w-full shrink-0 rounded-md border border-theme-border-moderate 2xl:block",s),...r})}function u(){return e.jsx("div",{className:"flex aspect-square h-5 w-5 shrink-0 items-center justify-center rounded-rounded bg-warning-400",children:e.jsx(c,{className:"h-3 w-3 shrink-0 fill-white"})})}export{g as P,u as T,j as a,w as b,b as c,N as d,y as e};
