"""Release [0.36.0].

Revision ID: 0.36.0
Revises: 9d8020441014
Create Date: 2023-03-20 11:29:20.384139

"""

# revision identifiers, used by Alembic.
revision = "0.36.0"
down_revision = "9d8020441014"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
