import{j as e}from"./@radix-C7hRs6Kx.js";function s({children:t}){return e.jsx("div",{className:"flex h-7 w-7 items-center justify-center rounded-sm bg-primary-100 text-text-lg font-semibold text-theme-text-brand",children:t})}export{s as N};
