"""Release [0.22.0].

Revision ID: 0.22.0
Revises: 8a64fbfecda0
Create Date: 2022-11-18 13:39:38.689448

"""

# revision identifiers, used by Alembic.
revision = "0.22.0"
down_revision = "8a64fbfecda0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
