"""Release [0.58.2].

Revision ID: 0.58.2
Revises: 0.58.1
Create Date: 2024-06-10 14:36:10.449091

"""

# revision identifiers, used by Alembic.
revision = "0.58.2"
down_revision = "0.58.1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
