"""Release [0.82.0].

Revision ID: 0.82.0
Revises: 0.81.0
Create Date: 2025-04-30 14:10:24.569601

"""

# revision identifiers, used by Alembic.
revision = "0.82.0"
down_revision = "0.81.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
