import{j as e,r as d}from"./@radix-C7hRs6Kx.js";import{S as h}from"./chevron-right-double-uNWbJT-C.js";import{m as l,aS as u,h as x,a6 as b,a7 as f,a8 as C,a9 as j}from"./index-D2iSHVZq.js";import{S as g}from"./chevron-down-A-rmltmI.js";function w({className:s,children:r,...a}){return e.jsxs("div",{className:l("flex h-9 items-center border-b border-theme-border-moderate bg-theme-surface-primary px-4 py-3",s),...a,children:[e.jsx(u,{asChild:!0,children:e.jsxs(x,{type:"button",className:"flex aspect-square items-center justify-center p-0",intent:"secondary",emphasis:"minimal",children:[e.jsx(h,{className:"h-5 w-5 fill-neutral-500"}),e.jsx("span",{className:"sr-only",children:"Close"})]})}),r]})}function D({title:s,children:r,initialOpen:a=!1,className:n,contentClassName:o,intent:i="primary",headerClassName:m,headerChildren:p}){const[t,c]=d.useState(a);return e.jsxs(b,{className:n,open:t,onOpenChange:c,children:[e.jsxs(f,{intent:i,className:m,children:[e.jsxs(C,{className:"flex w-full items-center gap-[10px]",children:[e.jsx(g,{className:` ${t?"":"-rotate-90"} h-5 w-5 rounded-md fill-neutral-500 transition-transform duration-200 hover:bg-neutral-200`}),s]}),p]}),e.jsx(j,{className:l("space-y-3 border-t border-theme-border-moderate bg-theme-surface-primary px-5 py-3",o),children:r})]})}export{D as C,w as S};
