"""Release [0.45.0].

Revision ID: 0.45.0
Revises: f3b3964e3a0f
Create Date: 2023-10-11 15:33:39.664836

"""

# revision identifiers, used by Alembic.
revision = "0.45.0"
down_revision = "f3b3964e3a0f"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
