"""Release [0.39.0].

Revision ID: 0.39.0
Revises: 743ec82b1b3c
Create Date: 2023-05-09 17:22:10.658453

"""

# revision identifiers, used by Alembic.
revision = "0.39.0"
down_revision = "743ec82b1b3c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
