"""Release [0.55.0].

Revision ID: 0.55.0
Revises: 4d688d8f7aff
Create Date: 2024-01-22 18:42:47.785139

"""

# revision identifiers, used by Alembic.
revision = "0.55.0"
down_revision = "4d688d8f7aff"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
