import{j as s}from"./@radix-C7hRs6Kx.js";import{I as a,h as p}from"./index-D2iSHVZq.js";import{u as w}from"./index.esm-D7jFlf5N.js";import{t as h}from"./zod-C0xYeTvL.js";import{P as u}from"./PasswordChecker-lYTOtNom.js";import{g as f}from"./form-schemas-CbvoEUHr.js";function b({displayUsername:r=!1,submitHandler:l,headline:o="Create your password",subHeadine:c="Select a password for your account"}){var t;const{register:e,watch:m,handleSubmit:d,formState:{isValid:i,errors:n}}=w({resolver:h(f(r)),mode:"onChange",criteriaMode:"all",defaultValues:{username:""}}),x=(t=n.newPassword)==null?void 0:t.types;return s.jsxs("div",{className:"space-y-5",children:[s.jsxs("div",{children:[s.jsx("h1",{className:"text-display-xs font-semibold",children:o}),s.jsx("p",{className:"text-theme-text-secondary",children:c})]}),s.jsxs("form",{onSubmit:d(l),className:"space-y-5",children:[s.jsx("div",{className:"space-y-2",children:r&&s.jsxs("div",{className:"space-y-0.5",children:[s.jsx("label",{className:"text-text-sm",children:"Username"}),s.jsx(a,{...e("username"),className:"w-full"})]})}),s.jsxs("div",{className:"space-y-0.5",children:[s.jsx("label",{className:"text-text-sm",children:"New Password"}),s.jsx(a,{...e("newPassword"),type:"password",className:"w-full"})]}),s.jsxs("div",{className:"space-y-0.5",children:[s.jsx("label",{className:"text-text-sm",children:"Confirm Password"}),s.jsx(a,{...e("confirmPassword"),type:"password",className:"w-full"})]}),s.jsx(u,{val:m("newPassword"),errors:x||[]}),s.jsx(p,{disabled:!i,className:"w-full text-center",size:"md",children:s.jsx("span",{className:"w-full",children:"Continue"})})]})]})}export{b as S};
